# ==================================================================================================================== #
#            __     __            _             _                                                                      #
#  _ __  _   \ \   / /__ _ __ ___(_) ___  _ __ (_)_ __   __ _                                                          #
# | '_ \| | | \ \ / / _ \ '__/ __| |/ _ \| '_ \| | '_ \ / _` |                                                         #
# | |_) | |_| |\ V /  __/ |  \__ \ | (_) | | | | | | | | (_| |                                                         #
# | .__/ \__, | \_/ \___|_|  |___/_|\___/|_| |_|_|_| |_|\__, |                                                         #
# |_|    |___/                                          |___/                                                          #
# ==================================================================================================================== #
# Authors:                                                                                                             #
#   Patrick Lehmann                                                                                                    #
#                                                                                                                      #
# License:                                                                                                             #
# ==================================================================================================================== #
# Copyright 2020-2023 Patrick Lehmann - Bötzingen, Germany                                                             #
#                                                                                                                      #
# Licensed under the Apache License, Version 2.0 (the "License");                                                      #
# you may not use this file except in compliance with the License.                                                     #
# You may obtain a copy of the License at                                                                              #
#                                                                                                                      #
#   http://www.apache.org/licenses/LICENSE-2.0                                                                         #
#                                                                                                                      #
# Unless required by applicable law or agreed to in writing, software                                                  #
# distributed under the License is distributed on an "AS IS" BASIS,                                                    #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                             #
# See the License for the specific language governing permissions and                                                  #
# limitations under the License.                                                                                       #
#                                                                                                                      #
# SPDX-License-Identifier: Apache-2.0                                                                                  #
# ==================================================================================================================== #
#
__author__ =    "Patrick Lehmann"
__email__ =     "Paebbels@gmail.com"
__copyright__ = "2020-2023, Patrick Lehmann"
__license__ =   "Apache License, Version 2.0"
__version__ =   "0.10.0"
__keywords__ =  ["Python3", "Template", "Versioning", "Git"]

from dataclasses  import dataclass, make_dataclass, field
from datetime     import date, time, datetime
from enum         import Enum, auto
from os           import environ
from pathlib      import Path
from subprocess   import run as subprocess_run, PIPE
from typing       import Union, Any, Dict

from pyTooling.Decorators import export
from pyTooling.Versioning       import SemVersion
from pyTooling.TerminalUI       import ILineTerminal

from pyVersioning.Utils import SelfDescriptive, GitHelper, GitShowCommand
from pyVersioning.AppVeyor      import AppVeyor
from pyVersioning.CIService     import WorkStation
from pyVersioning.Configuration import Configuration
from pyVersioning.GitLab        import GitLab
from pyVersioning.GitHub        import GitHub
from pyVersioning.Travis        import Travis


@export
class Tool(SelfDescriptive):
	"""This data structure class describes the tool name and version of pyVersioning."""

	_name:    str
	_version: SemVersion

	_public = ["name", "version"]

	def __init__(self, name: str, version: SemVersion):
		self._name = name
		self._version = version

	@property
	def name(self) -> str:
		return self._name

	@property
	def version(self) -> SemVersion:
		return self._version

	def __str__(self) -> str:
		return f"{self._name} {self._version}"


@export
class Date(date, SelfDescriptive):
	_public = ["day", "month", "year"]


@export
class Time(time, SelfDescriptive):
	_public = ["hour", "minute", "second"]


@export
class Person(SelfDescriptive):
	"""This data structure class describes an author with name and email address."""

	_name:  str
	_email: str

	_public = ["name", "email"]

	def __init__(self, name: str, email: str):
		self._name = name
		self._email = email

	@property
	def name(self) -> str:
		return self._name

	@property
	def email(self) -> str:
		return self._email

	def __str__(self) -> str:
		return f"{self._name} <{self._email}>"


@export
@dataclass
class Commit(SelfDescriptive):
	hash:      str
	date:      Date
	time:      Time
	author:    Person
	committer: Person
	comment:   str
	oneline:   str = field(init=False)

	_public = ["hash", "date", "time", "author", "committer", "comment", "oneline"]

	def __post_init__(self) -> None:
		"""Calculate `oneline` from `comment`."""

		if self.comment != "":
			self.oneline = self.comment.split("\n")[0]


@export
@dataclass
class Git(SelfDescriptive):
	commit:     Commit
	reference:  str = field(init=False)
	tag:        str = ""
	branch:     str = ""
	repository: str = ""

	_public = ["commit", "reference", "tag", "branch", "repository"]

	def __post_init__(self) -> None:
		"""Calculate `reference` from `tag` or `branch`."""

		if self.tag != "":
			self.reference = self.tag
		elif self.branch != "":
			self.reference = self.branch
		else:
			self.reference = "[Detached HEAD]"


@export
class Project(SelfDescriptive):
	_name:     str
	_variant:  str
	_version:  SemVersion

	_public = ["name", "variant", "version"]

	def __init__(self, name: str, version: Union[str, SemVersion] = None, variant: str = None) -> None:
		"""Assign fields and convert version string to a `Version` object."""

		self._name    = name    if name    is not None else ""
		self._variant = variant if variant is not None else ""

		if isinstance(version, SemVersion):
			self._version = version
		elif isinstance(version, str):
			self._version = SemVersion(version)
		elif version is None:
			self._version = SemVersion(0, 0, 0)

	@property
	def name(self) -> str:
		return self._name

	@property
	def variant(self) -> str:
		return self._variant

	@property
	def version(self) -> SemVersion:
		return self._version

	def __str__(self) -> str:
		return f"{self._name} - {self._variant} {self._version}"


@export
class Compiler(SelfDescriptive):
	_name:          str
	_version:       SemVersion
	_configuration: str
	_options:       str

	_public = ["name", "version", "configuration", "options"]

	def __init__(self, name: str, version: Union[str, SemVersion] = "", configuration: str = "", options: str = "") -> None:
		"""Assign fields and convert version string to a `Version` object."""

		self._name          = name          if name is not None else ""
		self._configuration = configuration if configuration is not None else ""
		self._options       = options       if options       is not None else ""

		if isinstance(version, SemVersion):
			self._version = version
		elif isinstance(version, str):
			self._version = SemVersion(version)
		elif version is None:
			self._version = SemVersion(0, 0, 0)

	@property
	def name(self) -> str:
		return self._name

	@property
	def version(self) -> SemVersion:
		return self._version

	@property
	def configuration(self) -> str:
		return self._configuration

	@property
	def options(self) -> str:
		return self._options


@export
class Build(SelfDescriptive):
	_date:     Date
	_time:     Time
	_compiler: Compiler

	_public = ["date", "time", "compiler"]

	def __init__(self, date: Date, time: Time, compiler: Compiler):
		self._date = date
		self._time = time
		self._compiler = compiler

	@property
	def date(self) -> Date:
		return self._date

	@property
	def time(self) -> Time:
		return self._time

	@property
	def compiler(self) -> Compiler:
		return self._compiler


@export
class Platforms(Enum):
	Workstation = auto()
	AppVeyor =    auto()
	GitHub =      auto()
	GitLab =      auto()
	Travis =      auto()


@export
class Versioning(ILineTerminal, GitHelper):
	platform:  Platforms = Platforms.Workstation
	variables: Dict

	def __init__(self, terminal: ILineTerminal):
		super().__init__(terminal)

		self.variables = {}

		if "APPVEYOR" in environ:
			self.platform = Platforms.AppVeyor
		elif "GITHUB_ACTIONS" in environ:
			self.platform = Platforms.GitHub
		elif "GITLAB_CI" in environ:
			self.platform = Platforms.GitLab
		elif "TRAVIS" in environ:
			self.platform = Platforms.Travis
		else:
			self.platform = Platforms.Workstation

	def loadDataFromConfiguration(self, config: Configuration) -> None:
		"""Preload versioning information from configuration file."""

		self.variables["project"] = self.getProject(config.project)
		self.variables["build"]   = self.getBuild(config.build)
		self.variables["version"] = self.getVersion(config.project)

	def collectData(self) -> None:
		"""Collect versioning information from environment including CI services (if available)."""

		if self.platform is Platforms.AppVeyor:
			self.service                = AppVeyor()
			self.variables["appveyor"]  = self.service.getEnvironment()
		elif self.platform is Platforms.GitHub:
			self.service                = GitHub()
			self.variables["github"]    = self.service.getEnvironment()
		elif self.platform is Platforms.GitLab:
			self.service                = GitLab()
			self.variables["gitlab"]    = self.service.getEnvironment()
		elif self.platform is Platforms.Travis:
			self.service                = Travis()
			self.variables["travis"]    = self.service.getEnvironment()
		else:
			self.service                = WorkStation()

		self.variables["tool"]     = Tool("pyVersioning", SemVersion(__version__))
		self.variables["git"]      = self.getGitInformation()
		self.variables["env"]      = self.getEnvironment()
		self.variables["platform"] = self.service.getPlatform()

		self.calculateData()

	def calculateData(self) -> None:
		if self.variables["git"].tag != "":
			pass

	def getVersion(self, config: Configuration.Project) -> SemVersion:
		if config.version is not None:
			return config.version
		else:
			return SemVersion("0.0.0")

	def getGitInformation(self) -> Git:
		return Git(
			commit=self.getLastCommit(),
			tag=self.getGitTag(),
			branch=self.getGitLocalBranch(),
			repository=self.getGitRemoteURL()
		)

	def getLastCommit(self) -> Commit:
		dt = self.getCommitDate()

		return Commit(
			hash=self.getGitHash(),
			date=dt.date(),
			time=dt.time(),
			author=self.getCommitAuthor(),
			committer=self.getCommitCommitter(),
			comment=self.getCommitComment()
		)

	def getGitHash(self) -> str:
		if self.platform is not Platforms.Workstation:
			return self.service.getGitHash()

		return self.execGitShow(GitShowCommand.CommitHash)

	def getCommitDate(self) -> datetime:
		if self.platform is not Platforms.Workstation:
			return self.service.getCommitDate()

		datetimeString = self.execGitShow(GitShowCommand.CommitDateTime)
		return datetime.fromtimestamp(int(datetimeString))

	def getCommitAuthor(self) -> Person:
		return Person(
			name=self.getCommitAuthorName(),
			email=self.getCommitAuthorEmail()
		)

	def getCommitAuthorName(self) -> str:
		return self.execGitShow(GitShowCommand.CommitAuthorName)

	def getCommitAuthorEmail(self) -> str:
		return self.execGitShow(GitShowCommand.CommitAuthorEmail)

	def getCommitCommitter(self) -> Person:
		return Person(
			name=self.getCommitCommitterName(),
			email=self.getCommitCommitterEmail()
		)

	def getCommitCommitterName(self) -> str:
		return self.execGitShow(GitShowCommand.CommitCommitterName)

	def getCommitCommitterEmail(self) -> str:
		return self.execGitShow(GitShowCommand.CommitCommitterEmail)

	def getCommitComment(self) -> str:
		return self.execGitShow(GitShowCommand.CommitComment)

	def getGitLocalBranch(self) -> str:
		if self.platform is not Platforms.Workstation:
			return self.service.getGitBranch()

		command = "git"
		arguments = ("branch", "--show-current")
		try:
			completed = subprocess_run((command, *arguments), stdout=PIPE, stderr=PIPE)
		except:
			return ""

		if completed.returncode == 0:
			return completed.stdout.decode("utf-8").split("\n")[0]
		else:
			message = completed.stderr.decode("utf-8")
			self.WriteFatal(f"Message from '{command}': {message}")

	def getGitRemoteBranch(self, localBranch: str = None) -> str:
		if self.platform is not Platforms.Workstation:
			return self.service.getGitBranch()

		if localBranch is None:
			localBranch = self.getGitLocalBranch()

		command = "git"
		arguments = ("config", f"branch.{localBranch}.merge")
		try:
			completed = subprocess_run((command, *arguments), stdout=PIPE, stderr=PIPE)
		except:
			raise Exception()

		if completed.returncode == 0:
			return completed.stdout.decode("utf-8").split("\n")[0]
		else:
			message = completed.stderr.decode("utf-8")
			self.WriteFatal(f"Message from '{command}': {message}")
			raise Exception

	def getGitRemote(self, localBranch: str = None) -> str:
		if localBranch is None:
			localBranch = self.getGitLocalBranch()

		command = "git"
		arguments = ("config", f"branch.{localBranch}.remote")
		try:
			completed = subprocess_run((command, *arguments), stdout=PIPE, stderr=PIPE)
		except:
			raise Exception

		if completed.returncode == 0:
			return completed.stdout.decode("utf-8").split("\n")[0]
		elif completed.returncode == 1:
			self.WriteWarning(f"Branch '{localBranch}' is not pushed to a remote.")
			return f"(local) {localBranch}"
		else:
			message = completed.stderr.decode("utf-8")
			self.WriteFatal(f"Message from '{command}': {message}")
			raise Exception

	def getGitTag(self) -> str:
		if self.platform is not Platforms.Workstation:
			return self.service.getGitTag()

		command = "git"
		arguments = ("tag", "--points-at", "HEAD")
		try:
			completed = subprocess_run((command, *arguments), stdout=PIPE, stderr=PIPE)
		except:
			raise Exception

		if completed.returncode == 0:
			return completed.stdout.decode("utf-8").split("\n")[0]
		else:
			message = completed.stderr.decode("utf-8")
			self.WriteFatal(f"Message from '{command}': {message}")
			raise Exception

	def getGitRemoteURL(self, remote: str = None) -> str:
		if self.platform is not Platforms.Workstation:
			return self.service.getGitRepository()

		if remote is None:
			remote = self.getGitRemote()

		command = "git"
		arguments = ("config", f"remote.{remote}.url")
		try:
			completed = subprocess_run((command, *arguments), stdout=PIPE, stderr=PIPE)
		except:
			raise Exception

		if completed.returncode == 0:
			return completed.stdout.decode("utf-8").split("\n")[0]
		else:
			message = completed.stderr.decode("utf-8")
			self.WriteFatal(f"Message from '{command}': {message}")
			raise Exception

	# 		self.WriteFatal(f"Message from '{command}': {message}")

	def getProject(self, config: Configuration.Project) -> Project:
		return Project(
			name=config.name,
			version=config.version,
			variant=config.variant
		)

	def getBuild(self, config: Configuration.Build) -> Build:
		dt = datetime.now()
		return Build(
			date=dt.date(),
			time=dt.time(),
			compiler=self.getCompiler(config.compiler)
		)

	def getCompiler(self, config: Configuration.Build.Compiler) -> Compiler:
		return Compiler(
			name=config.name,
			version=SemVersion(config.version),
			configuration=config.configuration,
			options=config.options
		)

	def getEnvironment(self) -> Any:
		env = {}
		for key, value in environ.items():
			if not key.isidentifier():
				self.WriteWarning(f"Skipping environment variable '{key}', because it's not a valid Python identifier.")
				continue
			key = key.replace("(", "_")
			key = key.replace(")", "_")
			key = key.replace(" ", "_")
			env[key] = value

		def func(s):
			for e in env.keys():
				yield (e, s.__getattribute__(e))

		Environment = make_dataclass(
			"Environment",
			[(name, str) for name in env.keys()],
#			bases=(SelfDescriptive,),
			namespace={
				"as_dict":       lambda self: env,
				"Keys":          lambda self: env.keys(),
				"KeyValuePairs": lambda self: func(self)
			},
			repr=True
		)

		return Environment(**env)

	def writeSourceFile(self, template: Path, filename: Path) -> None:
		content = template.read_text()

		# apply variables
		content = content.format(**self.variables)

		with filename.open("w") as file:
			file.write(content)



# ==================================================================================================================== #
#            __     __            _             _                                                                      #
#  _ __  _   \ \   / /__ _ __ ___(_) ___  _ __ (_)_ __   __ _                                                          #
# | '_ \| | | \ \ / / _ \ '__/ __| |/ _ \| '_ \| | '_ \ / _` |                                                         #
# | |_) | |_| |\ V /  __/ |  \__ \ | (_) | | | | | | | | (_| |                                                         #
# | .__/ \__, | \_/ \___|_|  |___/_|\___/|_| |_|_|_| |_|\__, |                                                         #
# |_|    |___/                                          |___/                                                          #
# ==================================================================================================================== #
# Authors:                                                                                                             #
#   Patrick Lehmann                                                                                                    #
#                                                                                                                      #
# License:                                                                                                             #
# ==================================================================================================================== #
# Copyright 2020-2023 Patrick Lehmann - Bötzingen, Germany                                                             #
#                                                                                                                      #
# Licensed under the Apache License, Version 2.0 (the "License");                                                      #
# you may not use this file except in compliance with the License.                                                     #
# You may obtain a copy of the License at                                                                              #
#                                                                                                                      #
#   http://www.apache.org/licenses/LICENSE-2.0                                                                         #
#                                                                                                                      #
# Unless required by applicable law or agreed to in writing, software                                                  #
# distributed under the License is distributed on an "AS IS" BASIS,                                                    #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                             #
# See the License for the specific language governing permissions and                                                  #
# limitations under the License.                                                                                       #
#                                                                                                                      #
# SPDX-License-Identifier: Apache-2.0                                                                                  #
# ==================================================================================================================== #
#
from argparse     import RawDescriptionHelpFormatter
from pathlib      import Path
from textwrap     import dedent
from typing       import NoReturn

from pyTooling.Decorators import export
from pyTooling.TerminalUI            import LineTerminal, Severity
from pyAttributes                    import Attribute
from pyAttributes.ArgParseAttributes import ArgParseMixin, CommandAttribute, ArgumentAttribute, DefaultAttribute

from pyVersioning                    import Versioning, Platforms, Project, SelfDescriptive
from pyVersioning.Configuration      import Configuration


@export
class ProjectAttributeGroup(Attribute):
	def __call__(self, func):
		self._AppendAttribute(func, ArgumentAttribute("--project-name",    metavar="<Name>",    dest="ProjectName",    type=str, help="Name of the project."))
		self._AppendAttribute(func, ArgumentAttribute("--project-variant", metavar="<Variant>", dest="ProjectVariant", type=str, help="Variant of the project."))
		self._AppendAttribute(func, ArgumentAttribute("--project-version", metavar="<Version>", dest="ProjectVersion", type=str, help="Version of the project."))
		return func


@export
class CompilerAttributeGroup(Attribute):
	def __call__(self, func):
		self._AppendAttribute(func, ArgumentAttribute("--compiler-name",    metavar="<Name>",    dest="CompilerName",    type=str, help="Used compiler."))
		self._AppendAttribute(func, ArgumentAttribute("--compiler-version", metavar="<Version>", dest="CompilerVersion", type=str, help="Used compiler version."))
		self._AppendAttribute(func, ArgumentAttribute("--compiler-config",  metavar="<Config>",  dest="CompilerConfig",  type=str, help="Used compiler configuration."))
		self._AppendAttribute(func, ArgumentAttribute("--compiler-options", metavar="<Options>", dest="CompilerOptions", type=str, help="Used compiler options."))
		return func


@export
class Application(LineTerminal, ArgParseMixin):
	HeadLine:     str = "Version file generator."

	__configFile: Path
	_config:      Configuration
	_versioning:  Versioning

	def __init__(self) -> None:
		super().__init__()

		ArgParseMixin.__init__(
			self,
			description=self.HeadLine,
			formatter_class=RawDescriptionHelpFormatter,
			add_help=False
		)

		self._LOG_MESSAGE_FORMAT__[Severity.Fatal] =   "{DARK_RED}[FATAL] {message}{NOCOLOR}"
		self._LOG_MESSAGE_FORMAT__[Severity.Error] =   "{RED}[ERROR] {message}{NOCOLOR}"
		self._LOG_MESSAGE_FORMAT__[Severity.Warning] = "{YELLOW}[WARNING] {message}{NOCOLOR}"
		self._LOG_MESSAGE_FORMAT__[Severity.Normal]=   "{GRAY}{message}{NOCOLOR}"

	def Initialize(self) -> None:
		if not self.__configFile.exists():
			self.WriteWarning(f"Configuration file '{self.__configFile}' does not exist.")
			self._config = Configuration()
		else:
			self._config = Configuration(self.__configFile)

		self._versioning = Versioning(self)
		self._versioning.loadDataFromConfiguration(self._config)
		self._versioning.collectData()

	def PrintHeadline(self) -> None:
		self.WriteNormal("{HEADLINE}{line}".format(line="=" * 80, **LineTerminal.Foreground))
		self.WriteNormal("{HEADLINE}{headline: ^80s}".format(headline=self.HeadLine, **LineTerminal.Foreground))
		self.WriteNormal("{HEADLINE}{line}".format(line="=" * 80, **LineTerminal.Foreground))

	def Run(self, configFile: Path) -> NoReturn:
		self.__configFile = configFile

		super().Run(self)
		self.exit()

	@DefaultAttribute()
	def HandleDefault(self, args) -> None:
		self.PrintHeadline()
		self.MainParser.print_help()

	@CommandAttribute("help", help="Display help page(s) for the given command name.")
	@ArgumentAttribute(metavar="Command", dest="Command", type=str, nargs="?", help="Print help page(s) for a command.")
	def HandleHelp(self, args) -> None:
		self.PrintHeadline()

		if args.Command is None:
			self.MainParser.print_help()
		elif args.Command == "help":
			self.WriteError("This is a recursion ...")
		else:
			try:
				self.SubParsers[args.Command].print_help()
			except KeyError:
				self.WriteError(f"Command {args.Command} is unknown.")

	@CommandAttribute("fillout", help="Read a template and replace tokens with version information.")
	@ProjectAttributeGroup()
	@CompilerAttributeGroup()
	@ArgumentAttribute(metavar="<Template file>", dest="Template", type=str, help="Template input filename.")
	@ArgumentAttribute(metavar="<Output file>",   dest="Filename", type=str, help="Output filename.")
	def HandleFillOut(self, args) -> None:
		self.PrintHeadline()
		self.Initialize()

		templateFile = Path(args.Template)
		if not templateFile.exists():
			self.WriteError(f"Template file '{templateFile}' does not exist.")

		outputFile = Path(args.Filename)
		if not outputFile.parent.exists():
			self.WriteWarning(f"Directory for file '{outputFile}' does not exist. Directory will be created")
			try:
				outputFile.parent.mkdir()
			except:
				self.WriteError(f"Failed to create the directory '{outputFile.parent}' for the output file.")
		elif outputFile.exists():
			self.WriteWarning(f"Output file '{outputFile}' already exists. This file will be overwritten.")

		self.ExitOnPreviousErrors()

		self.UpdateProject(args)
		self.UpdateCompiler(args)

		self._versioning.writeSourceFile(templateFile, outputFile)

	@CommandAttribute("variables", help="Print all available variables.")
	@ProjectAttributeGroup()
	@CompilerAttributeGroup()
	def HandleVariables(self, args) -> None:
		self.PrintHeadline()
		self.Initialize()

		self.UpdateProject(args)
		self.UpdateCompiler(args)

		def print(key, value, indent):
			key = ("  " * indent) + str(key)
			self.WriteNormal(f"{key:24}: {value!s}")
			if isinstance(value, SelfDescriptive):
				for k, v in value.KeyValuePairs():
					print(k, v, indent + 1)

		for key,value in self._versioning.variables.items():
			print(key, value, 0)

	@CommandAttribute("json", help="Write all available variables as JSON.")
	@ProjectAttributeGroup()
	@CompilerAttributeGroup()
	@ArgumentAttribute(metavar="<Output file>",   dest="Filename", type=str, nargs="?", help="Output filename.")
	def HandleJSON(self, args) -> None:
		self.Initialize()

		self.UpdateProject(args)
		self.UpdateCompiler(args)

		content = dedent("""\
		{{
		  version: {{
		    major: {version.major},
		    minor: {version.minor},
		    patch: {version.patch},
		   flags: {version.flags}
		  }}
		}}
		""")
		output = content.format(**self._versioning.variables)
		self.WriteNormal(output)

	@CommandAttribute("yaml", help="Write all available variables as YAML.")
	@ProjectAttributeGroup()
	@CompilerAttributeGroup()
	@ArgumentAttribute(metavar="<Output file>",   dest="Filename", type=str, nargs="?", help="Output filename.")
	def HandleYAML(self, args) -> None:
		self.Initialize()

		self.UpdateProject(args)
		self.UpdateCompiler(args)

		yamlEnvironment = "\n"
		# for key, value in self._versioning.variables["env"].as_dict().items():
		# 	yamlEnvironment += f"    {key}: {value}\n"

		yamlAppVeyor = "\n#   not found"
		yamlGitHub =   "\n#   not found"
		yamlGitLab =   "\n#   not found"
		yamlTravis =   "\n#   not found"
		if self._versioning.platform is Platforms.AppVeyor:
			yamlAppVeyor = "\n"
			for key, value in self._versioning.variables["appveyor"].as_dict().items():
				yamlAppVeyor += f"    {key}: {value}\n"
		elif self._versioning.platform is Platforms.GitHub:
			yamlGitHub = "\n"
			for key, value in self._versioning.variables["github"].as_dict().items():
				yamlGitHub += f"    {key}: {value}\n"
		elif self._versioning.platform is Platforms.GitLab:
			yamlGitLab = "\n"
			for key, value in self._versioning.variables["gitlab"].as_dict().items():
				yamlGitLab += f"    {key}: {value}\n"
		elif self._versioning.platform is Platforms.Travis:
			yamlTravis = "\n"
			for key, value in self._versioning.variables["travis"].as_dict().items():
				yamlTravis += f"    {key}: {value}\n"

		content = dedent("""\
		  version: {version!s}
		    major: {version.major}
		    minor: {version.minor}
		    patch: {version.patch}
		    flags: {version.flags}

		  git:
		    commit:
		      hash: {git.commit.hash}
		      date: {git.commit.date}
		    reference: {git.reference}
		    branch: {git.branch}
		    tag: {git.tag}
		    repository: {git.repository}

		  project:
		    name: {project.name}
		    variant: {project.variant}

		  build:
		    date: {build.date}
		    compiler:
		      name: {build.compiler.name}
		      version: {build.compiler.version}
		      configuration: {build.compiler.configuration}
		      options: {build.compiler.options}

		  platform:
		    ci-service: {platform.ci_service}
		  appveyor: {yamlAppVeyor}
		  github: {yamlGitHub}
		  gitlab: {yamlGitLab}
		  travis: {yamlTravis}
		  env:{yamlEnvironment}
		""")
		output = content.format(**self._versioning.variables, yamlEnvironment=yamlEnvironment, yamlAppVeyor=yamlAppVeyor, yamlGitHub=yamlGitHub, yamlGitLab=yamlGitLab, yamlTravis=yamlTravis)
		self.WriteNormal(output)

	def UpdateProject(self, args) -> None:
		if "project" not in self._versioning.variables:
			self._versioning.variables["project"] = Project(args.ProjectName, args.ProjectVersion, args.ProjectVariant)
		elif args.ProjectName is not None:
			self._versioning.variables["project"]._name = args.ProjectName

		if args.ProjectVariant is not None:
			self._versioning.variables["project"]._variant = args.ProjectVariant

		if args.ProjectVersion is not None:
			self._versioning.variables["project"]._version = args.ProjectVersion

	def UpdateCompiler(self, args) -> None:
		if args.CompilerName is not None:
			self._versioning.variables["build"]._compiler._name = args.CompilerName
		if args.CompilerVersion is not None:
			self._versioning.variables["build"]._compiler._version = args.CompilerVersion
		if args.CompilerConfig is not None:
			self._versioning.variables["build"]._compiler._configuration = args.CompilerConfig
		if args.CompilerOptions is not None:
			self._versioning.variables["build"]._compiler._options = args.CompilerOptions


def main() -> NoReturn:
	configFile = Path(".pyVersioning.yml")

	Application.versionCheck((3, 7, 0))
	application = Application()
	application.Run(configFile)


if __name__ == "__main__":
	main()



# ==================================================================================================================== #
#            __     __            _             _                                                                      #
#  _ __  _   \ \   / /__ _ __ ___(_) ___  _ __ (_)_ __   __ _                                                          #
# | '_ \| | | \ \ / / _ \ '__/ __| |/ _ \| '_ \| | '_ \ / _` |                                                         #
# | |_) | |_| |\ V /  __/ |  \__ \ | (_) | | | | | | | | (_| |                                                         #
# | .__/ \__, | \_/ \___|_|  |___/_|\___/|_| |_|_|_| |_|\__, |                                                         #
# |_|    |___/                                          |___/                                                          #
# ==================================================================================================================== #
# Authors:                                                                                                             #
#   Patrick Lehmann                                                                                                    #
#                                                                                                                      #
# License:                                                                                                             #
# ==================================================================================================================== #
# Copyright 2020-2023 Patrick Lehmann - Bötzingen, Germany                                                             #
#                                                                                                                      #
# Licensed under the Apache License, Version 2.0 (the "License");                                                      #
# you may not use this file except in compliance with the License.                                                     #
# You may obtain a copy of the License at                                                                              #
#                                                                                                                      #
#   http://www.apache.org/licenses/LICENSE-2.0                                                                         #
#                                                                                                                      #
# Unless required by applicable law or agreed to in writing, software                                                  #
# distributed under the License is distributed on an "AS IS" BASIS,                                                    #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                             #
# See the License for the specific language governing permissions and                                                  #
# limitations under the License.                                                                                       #
#                                                                                                                      #
# SPDX-License-Identifier: Apache-2.0                                                                                  #
# ==================================================================================================================== #
#
"""Module for CI service base classes."""
from dataclasses  import make_dataclass
from datetime     import datetime
from os           import environ
from typing       import Dict, Optional as Nullable

from pyTooling.Decorators import export
from pyTooling.MetaClasses import ExtendedType, abstractmethod

from pyVersioning import SelfDescriptive, GitHelper, GitShowCommand


@export
class ServiceException(Exception):
	""".. todo:: ServiceException needs documentation"""


@export
class Platform(SelfDescriptive):
	""".. todo:: Platform needs documentation"""

	_ciService: str

	_public = ['ci_service']

	def __init__(self, ciService: str):
		self._ciService = ciService

	@property
	def ci_service(self) -> str:
		return self._ciService


@export
class BaseService(metaclass=ExtendedType):
	"""Base-class to collect platform and environment information from e.g. environment variables."""

	@abstractmethod
	def getPlatform(self) -> Platform:
		""".. todo:: getPlatform needs documentation"""


@export
class CIService(BaseService, GitHelper):
	"""Base-class to collect Git and other platform and environment information from CI service environment variables."""

	ENV_INCLUDE_FILTER = ()
	ENV_EXCLUDE_FILTER = ()
	ENV_INCLUDES =       []
	ENV_EXCLUDES =       []

	def getEnvironment(self) -> Dict[str, str]:
		""".. todo:: getEnvironment needs documentation"""

		filteredEnv = {key:value for (key,value) in environ.items() if key.startswith(self.ENV_INCLUDE_FILTER) and not key.endswith(self.ENV_EXCLUDE_FILTER)}

		# manually add some variables
		for key in self.ENV_INCLUDES:
			try:
				filteredEnv[key] = environ[key]
			except KeyError:
				pass

		# manually delete some variables
		for key in self.ENV_EXCLUDES:
			try:
				del filteredEnv[key]
			except KeyError:
				pass

		def func(s):
			for e in filteredEnv.keys():
				yield (e, s.__getattribute__(e))

		Environment = make_dataclass(
			"Environment",
			[(name, str) for name in filteredEnv.keys()],
			bases=(SelfDescriptive,),
			namespace={
				'as_dict':       lambda self: filteredEnv,
				'Keys':          lambda self: filteredEnv.keys(),
				'KeyValuePairs': lambda self: func(self)
			},
			repr=True
		)

		return Environment(**filteredEnv)

	@abstractmethod
	def getGitHash(self) -> str:
		""".. todo:: getGithash needs documentation"""

	# @abstractmethod
	def getCommitDate(self) -> datetime:
		""".. todo:: getCommitDate needs documentation"""

		datetimeString = self.execGitShow(GitShowCommand.CommitDateTime, self.getGitHash())
		return datetime.fromtimestamp(int(datetimeString))

	@abstractmethod
	def getGitBranch(self) -> Nullable[str]:
		""".. todo:: getGitBranch needs documentation"""

	@abstractmethod
	def getGitTag(self) -> Nullable[str]:
		""".. todo:: getGitTag needs documentation"""

	@abstractmethod
	def getGitRepository(self) -> str:
		""".. todo:: getGitRepository needs documentation"""


@export
class WorkStation(BaseService):
	def getPlatform(self) -> Platform:
		return Platform("NO-CI")



# ==================================================================================================================== #
#            __     __            _             _                                                                      #
#  _ __  _   \ \   / /__ _ __ ___(_) ___  _ __ (_)_ __   __ _                                                          #
# | '_ \| | | \ \ / / _ \ '__/ __| |/ _ \| '_ \| | '_ \ / _` |                                                         #
# | |_) | |_| |\ V /  __/ |  \__ \ | (_) | | | | | | | | (_| |                                                         #
# | .__/ \__, | \_/ \___|_|  |___/_|\___/|_| |_|_|_| |_|\__, |                                                         #
# |_|    |___/                                          |___/                                                          #
# ==================================================================================================================== #
# Authors:                                                                                                             #
#   Patrick Lehmann                                                                                                    #
#                                                                                                                      #
# License:                                                                                                             #
# ==================================================================================================================== #
# Copyright 2020-2023 Patrick Lehmann - Bötzingen, Germany                                                             #
#                                                                                                                      #
# Licensed under the Apache License, Version 2.0 (the "License");                                                      #
# you may not use this file except in compliance with the License.                                                     #
# You may obtain a copy of the License at                                                                              #
#                                                                                                                      #
#   http://www.apache.org/licenses/LICENSE-2.0                                                                         #
#                                                                                                                      #
# Unless required by applicable law or agreed to in writing, software                                                  #
# distributed under the License is distributed on an "AS IS" BASIS,                                                    #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                             #
# See the License for the specific language governing permissions and                                                  #
# limitations under the License.                                                                                       #
#                                                                                                                      #
# SPDX-License-Identifier: Apache-2.0                                                                                  #
# ==================================================================================================================== #
#
from enum       import Enum, auto
from subprocess import run as subprocess_run, PIPE
from typing     import List

from pyTooling.Decorators  import export
from pyTooling.MetaClasses import ExtendedType


@export
class SelfDescriptive(metaclass=ExtendedType):

	# TODO: could this be filled with a decorator?
	_public: List[str]

	def Keys(self):
		for element in self._public:
			yield element

	def KeyValuePairs(self):
		for element in self._public:
			value = self.__getattribute__(element)
			yield (element, value)


@export
class GitShowCommand(Enum):
	CommitDateTime =       auto()
	CommitAuthorName =     auto()
	CommitAuthorEmail =    auto()
	CommitCommitterName =  auto()
	CommitCommitterEmail = auto()
	CommitHash =           auto()
	CommitComment =        auto()


@export
class ToolException(Exception):
	command:      str
	errorMessage: str

	def __init__(self, command: str, errorMessage: str):
		self.command = command
		self.errorMessage = errorMessage


@export
class GitHelper:
	__GIT_SHOW_COMMAND_TO_FORMAT_LOOKUP = {
		GitShowCommand.CommitHash: "%H",
		GitShowCommand.CommitDateTime: "%ct",
		GitShowCommand.CommitAuthorName: "%an",
		GitShowCommand.CommitAuthorEmail: "%ae",
		GitShowCommand.CommitCommitterName: "%cn",
		GitShowCommand.CommitCommitterEmail: "%ce",
		GitShowCommand.CommitComment: "%B",
	}

	def execGitShow(self, command: GitShowCommand, ref: str = "HEAD") -> str:
		format = f"--format='{self.__GIT_SHOW_COMMAND_TO_FORMAT_LOOKUP[command]}'"

		command = "git"
		arguments = ("show", "-s", format, ref)
		try:
			completed = subprocess_run((command, *arguments), stdout=PIPE, stderr=PIPE)
		except Exception as ex:
			raise ToolException(f"{command} {' '.join(arguments)}", str(ex))

		if completed.returncode == 0:
			comment = completed.stdout.decode("utf-8")
			return comment[1:-2]
		else:
			message = completed.stderr.decode("utf-8")
			raise ToolException(f"{command} {' '.join(arguments)}", message)



# ==================================================================================================================== #
#            __     __            _             _                                                                      #
#  _ __  _   \ \   / /__ _ __ ___(_) ___  _ __ (_)_ __   __ _                                                          #
# | '_ \| | | \ \ / / _ \ '__/ __| |/ _ \| '_ \| | '_ \ / _` |                                                         #
# | |_) | |_| |\ V /  __/ |  \__ \ | (_) | | | | | | | | (_| |                                                         #
# | .__/ \__, | \_/ \___|_|  |___/_|\___/|_| |_|_|_| |_|\__, |                                                         #
# |_|    |___/                                          |___/                                                          #
# ==================================================================================================================== #
# Authors:                                                                                                             #
#   Patrick Lehmann                                                                                                    #
#                                                                                                                      #
# License:                                                                                                             #
# ==================================================================================================================== #
# Copyright 2020-2023 Patrick Lehmann - Bötzingen, Germany                                                             #
#                                                                                                                      #
# Licensed under the Apache License, Version 2.0 (the "License");                                                      #
# you may not use this file except in compliance with the License.                                                     #
# You may obtain a copy of the License at                                                                              #
#                                                                                                                      #
#   http://www.apache.org/licenses/LICENSE-2.0                                                                         #
#                                                                                                                      #
# Unless required by applicable law or agreed to in writing, software                                                  #
# distributed under the License is distributed on an "AS IS" BASIS,                                                    #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                             #
# See the License for the specific language governing permissions and                                                  #
# limitations under the License.                                                                                       #
#                                                                                                                      #
# SPDX-License-Identifier: Apache-2.0                                                                                  #
# ==================================================================================================================== #
#
"""pyVersioning configuration file in YAML format."""
from pathlib               import Path
from typing                import Dict, Optional as Nullable

from ruamel.yaml           import YAML
from pyTooling.Decorators  import export
from pyTooling.MetaClasses import ExtendedType
from pyTooling.Versioning  import SemVersion


@export
class Base:
	root:   'Base'
	parent: Nullable['Base']

	def __init__(self, root: 'Base', parent: Nullable['Base']):
		self.root = root
		self.parent = parent


@export
class Configuration(Base, metaclass=ExtendedType):
	class Project(Base):
		name:    str
		variant: str
		version: SemVersion

		def __init__(self, root: 'Base', parent: 'Base', settings: Dict):
			super().__init__(root, parent)

			self.name =    settings["name"]
			self.variant = settings["variant"] if "variant" in settings else None
			self.version = SemVersion(settings["version"]) if "version" in settings else None

	class Build(Base):
		class Compiler(Base):
			name:          str
			version:       str
			configuration: str
			options:       str

			def __init__(self, root: 'Base', parent: 'Base', settings: Dict):
				super().__init__(root, parent)

				self.name =          settings["name"]
				self.version =       settings["version"]
				self.configuration = settings["configuration"]
				self.options =       settings["options"]

		compiler: Compiler

		def __init__(self, root: 'Base', parent: 'Base', settings: Dict):
			super().__init__(root, parent)

			self.compiler = self.Compiler(root, self, settings["compiler"]) if "compiler" in settings else None

	version: int
	project: Project
	build:   Build

	def __init__(self, configFile: Path = None):
		super().__init__(self, None)

		if configFile is None:
			self.version = 1
			self.project = self.Project(self, self, {
				"name":    "",
				"variant": "",
				"version": "v0.0.0"
			})
			self.build = self.Build(self, self, {
				"compiler": {
					"name": "",
					"version": "v0.0.0",
					"configuration": "",
					"options": ""
				}
			})
		else:
			self.load(configFile)

	def load(self, configFile: Path):
		yaml =   YAML()
		config = yaml.load(configFile)

		self.version = int(config["version"]) if "version" in config else 1

		if self.version == 1:
			self.loadVersion1(config)

	def loadVersion1(self, config):
		self.project = self.Project(self, self, config["project"]) if "project" in config else None
		self.build =   self.Build(self, self, config["build"])     if "build" in config   else None



# ==================================================================================================================== #
#            __     __            _             _                                                                      #
#  _ __  _   \ \   / /__ _ __ ___(_) ___  _ __ (_)_ __   __ _                                                          #
# | '_ \| | | \ \ / / _ \ '__/ __| |/ _ \| '_ \| | '_ \ / _` |                                                         #
# | |_) | |_| |\ V /  __/ |  \__ \ | (_) | | | | | | | | (_| |                                                         #
# | .__/ \__, | \_/ \___|_|  |___/_|\___/|_| |_|_|_| |_|\__, |                                                         #
# |_|    |___/                                          |___/                                                          #
# ==================================================================================================================== #
# Authors:                                                                                                             #
#   Patrick Lehmann                                                                                                    #
#                                                                                                                      #
# License:                                                                                                             #
# ==================================================================================================================== #
# Copyright 2020-2023 Patrick Lehmann - Bötzingen, Germany                                                             #
#                                                                                                                      #
# Licensed under the Apache License, Version 2.0 (the "License");                                                      #
# you may not use this file except in compliance with the License.                                                     #
# You may obtain a copy of the License at                                                                              #
#                                                                                                                      #
#   http://www.apache.org/licenses/LICENSE-2.0                                                                         #
#                                                                                                                      #
# Unless required by applicable law or agreed to in writing, software                                                  #
# distributed under the License is distributed on an "AS IS" BASIS,                                                    #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                             #
# See the License for the specific language governing permissions and                                                  #
# limitations under the License.                                                                                       #
#                                                                                                                      #
# SPDX-License-Identifier: Apache-2.0                                                                                  #
# ==================================================================================================================== #
#
"""Package installer for 'Write version information for any programming language as source file'."""
from pathlib             import Path
from pyTooling.Packaging import DescribePythonPackageHostedOnGitHub

gitHubNamespace =        "Paebbels"
packageName =            "pyVersioning"
packageDirectory =       packageName.replace(".", "/")
packageInformationFile = Path(f"{packageDirectory}/__init__.py")

DescribePythonPackageHostedOnGitHub(
	packageName=packageName,
	description="Write version information for any programming language as source file.",
	gitHubNamespace=gitHubNamespace,
	sourceFileWithVersion=packageInformationFile,
  consoleScripts={
    "pyVersioning": "pyVersioning.cli:main",
  }
)



# ==================================================================================================================== #
#            __     __            _             _                                                                      #
#  _ __  _   \ \   / /__ _ __ ___(_) ___  _ __ (_)_ __   __ _                                                          #
# | '_ \| | | \ \ / / _ \ '__/ __| |/ _ \| '_ \| | '_ \ / _` |                                                         #
# | |_) | |_| |\ V /  __/ |  \__ \ | (_) | | | | | | | | (_| |                                                         #
# | .__/ \__, | \_/ \___|_|  |___/_|\___/|_| |_|_|_| |_|\__, |                                                         #
# |_|    |___/                                          |___/                                                          #
# ==================================================================================================================== #
# Authors:                                                                                                             #
#   Patrick Lehmann                                                                                                    #
#                                                                                                                      #
# License:                                                                                                             #
# ==================================================================================================================== #
# Copyright 2020-2023 Patrick Lehmann - Bötzingen, Germany                                                             #
#                                                                                                                      #
# Licensed under the Apache License, Version 2.0 (the "License");                                                      #
# you may not use this file except in compliance with the License.                                                     #
# You may obtain a copy of the License at                                                                              #
#                                                                                                                      #
#   http://www.apache.org/licenses/LICENSE-2.0                                                                         #
#                                                                                                                      #
# Unless required by applicable law or agreed to in writing, software                                                  #
# distributed under the License is distributed on an "AS IS" BASIS,                                                    #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                             #
# See the License for the specific language governing permissions and                                                  #
# limitations under the License.                                                                                       #
#                                                                                                                      #
# SPDX-License-Identifier: Apache-2.0                                                                                  #
# ==================================================================================================================== #
#
"""GitHub specific code to collect the build environment."""
from datetime import datetime
from os     import environ
from typing import Optional as Nullable

from pyTooling.Decorators import export

from pyVersioning.CIService import CIService, Platform, ServiceException


@export
class GitHub(CIService):
	"""Collect Git and other platform and environment information from environment variables provided by GitHub Actions."""

	ENV_INCLUDE_FILTER =  ("GITHUB_", )
	ENV_EXCLUDE_FILTER =  ("_TOKEN", )
	ENV_INCLUDES =        ('CI', )
	ENV_EXCLUDES =        []

	def getPlatform(self) -> Platform:
		return Platform("github")

	def getGitHash(self) -> str:
		try:
			return environ['GITHUB_SHA']
		except KeyError as ex:
			raise ServiceException from ex

	def getGitBranch(self) -> Nullable[str]:
		branchPrefix = "refs/heads/"

		try:
			ref = environ['GITHUB_REF']
			if ref.startswith(branchPrefix):
				return ref[len(branchPrefix):]
		except KeyError:
			return None

	def getGitTag(self) -> Nullable[str]:
		tagPrefix    = "refs/tags/"

		try:
			ref = environ['GITHUB_REF']
			if ref.startswith(tagPrefix):
				return ref[len(tagPrefix):]
		except KeyError:
			return None

	def getGitRepository(self) -> str:
		try:
			return environ['GITHUB_REPOSITORY']
		except KeyError as ex:
			raise ServiceException from ex
